#pragma hdrfile listbox.csm
#include <windows.h>
#include <string>
#include <debug.h>
#pragma hdrstop
#include "select.rh"
#include "listbox.h"
#include "system.h"

FARPROC SelectorList::m_pListBoxHandler = 0;

SelectorList::SelectorList( HWND hwndParent,
                            const char* szTitle,
                            int configFlags,
                            int positionFlags,
                            const RECT& posRect ) : m_fShrunk( false ),
                                               m_CfgFlags( configFlags ),
                                               m_PositionFlags( positionFlags ),
                                               m_hwndParent( hwndParent ),
                                               m_title( szTitle ),
                                               m_hFont( 0 ),
                                               m_iChoice( 0xFFFF ),
                                               m_fHidden( false ),
                                               m_fClosed( false ),
                                               m_fDragging( false )
{
    CopyMemory( &m_fullSizeRect, &posRect, sizeof( RECT ));

    if( SmallFont() )
    {
        m_hFont = CreateFont(
        -MulDiv(8, GetDeviceCaps(GetDC(GetDesktopWindow()), LOGPIXELSY), 72),
        0,
        0,
        0,
        400,
        false,
        false,
        false,
        ANSI_CHARSET,
        OUT_DEFAULT_PRECIS,
        CLIP_DEFAULT_PRECIS,
        PROOF_QUALITY,
        FF_DONTCARE | DEFAULT_PITCH,
        "Arial" );
    }
}

SelectorList::~SelectorList()
{

    if( SmallFont() )
    {
        DeleteObject( m_hFont );
    }

    if( UseTaskBar() )
    {
        TrayMessage( m_hwnd, NIM_DELETE, NOTIFY1, 0, 0 );
    }
}

bool SelectorList::Create()
{
    // set styles
    DWORD dwLBStyle = LBS_STANDARD      |
                      LBS_HASSTRINGS    |
                      WS_CAPTION        |
                      WS_VISIBLE        |
                      WS_SYSMENU;

    if( !NoSizing() )
    {
        dwLBStyle |=  WS_SIZEBOX;
    }

    // smaller caption bar
    DWORD dwStyleEx = WS_EX_TOOLWINDOW |
                      WS_EX_CLIENTEDGE |
                      WS_EX_NOPARENTNOTIFY;

    if( AlwaysOnTop() )
    {
        dwStyleEx |= WS_EX_TOPMOST;
    }

    // store parent position to determine if it moved
    GetWindowRect( Parent(), &m_rectParent );

    CalcDimensions( m_fullSizeRect );

    m_hwnd = CreateWindowEx(  dwStyleEx,
                              "LISTBOX",
                              m_title.c_str(),
                              dwLBStyle,
                              m_fullSizeRect.left,
                              m_fullSizeRect.top,
                              m_fullSizeRect.right - m_fullSizeRect.left,
                              m_fullSizeRect.bottom - m_fullSizeRect.top,
                              GetDesktopWindow(),
                              0,
                              0,
                              0 );

    if( 0 == m_hwnd )
    {
        TRACE( "Error creating list box: " << GetLastError() );
        return false;
    }

    // subclass list box
    m_pListBoxHandler = (FARPROC) GetWindowLong( m_hwnd, GWL_WNDPROC );
    SetWindowLong( m_hwnd, GWL_WNDPROC, (LONG)SelectorList::CallBack );
    SetWindowLong( m_hwnd, GWL_USERDATA, (LONG) this );

    if( SmallFont() )
    {
        bool fRedraw = false;
        SendMessage( m_hwnd, WM_SETFONT, (WPARAM)m_hFont, MAKELPARAM(fRedraw, 0) );
    }


    if( ChangeIcon() || AlternateIcon() )
    {
        HICON hIcon;
        if( AlternateIcon() )
        {
            hIcon = LoadIcon( GetModuleHandle( "select" ), MAKEINTRESOURCE( IDI_TSE ));
        }
        else
        {
            hIcon = LoadIcon( 0, IDI_WINLOGO );
        }
        SendMessage( Parent(), WM_SETICON, (WPARAM)true, (LPARAM) hIcon );
    }

    if( UseTaskBar() )
    {
        HICON hIcon;
        hIcon = LoadIcon( GetModuleHandle( "select" ), MAKEINTRESOURCE( IDI_TSE ));
        TrayMessage( m_hwnd, NIM_ADD, NOTIFY1, hIcon, "TSE/32 Selector" );
    }

    ShowWindow( m_hwnd, SW_SHOW );
    UpdateWindow( m_hwnd );
    return true;
}

LRESULT FAR PASCAL SelectorList::CallBack( HWND hwnd,
                                           UINT message,
                                           WPARAM wParam,
                                           LPARAM lParam )
{
    switch( message )
    {
        case WM_DESTROY:
            TRACE( "Quitting app" );

            PostQuitMessage( 0 );
            break;
    }

    SelectorList* pList = (SelectorList*) GetWindowLong( hwnd, GWL_USERDATA );

    return pList->OnMessage( message, wParam, lParam );
}

LRESULT SelectorList::OnMessage( UINT message,
                                 WPARAM wParam,
                                 LPARAM lParam )
{
    switch( message )
    {
        case WM_LBUTTONDOWN:
        {
            CallWindowProc( m_pListBoxHandler,
                            m_hwnd,
                            message,
                            wParam,
                            lParam);

            OnLButtonDown( wParam );
            break;
        }

        case WM_LBUTTONDBLCLK:
        {
            CallWindowProc( m_pListBoxHandler,
                            m_hwnd,
                            message,
                            wParam,
                            lParam);

            OnLButtonDblClick( wParam );
            break;
        }

        case WM_CHAR:
        {
            OnChar( (char)wParam );
            break;
        }

        case WM_MBUTTONDOWN:
        case WM_NCMBUTTONDOWN:
        {
            if( ScreenSaver() )
            {
                SendMessage( Hwnd(), WM_SYSCOMMAND, SC_SCREENSAVE, 0 );
            }
            break;
        }

        case WM_RBUTTONDOWN:
        case WM_NCRBUTTONDOWN:
        {
            OnRButtonDown( wParam );
            break;
        }

        case WM_SYSCOMMAND:
        {
            TRACE( "Syscommand: " << hex << wParam );
            TRACE( "Syscommand (lParam): " << hex << lParam );
            ResetSearch();
            if( SC_CLOSE == wParam && !AllowClose() )
            {
                return 0;
            }

            if( SC_CLOSE == wParam && CloseTSE() )
            {
                m_fClosed = true;
            }

            // drag / move (undocumented - not in winuser.h)
            if( 0xF012 == wParam )
            {
                if( !OnDragging() )
                {
                    return 0;
                }

                m_fDragging = true;
            }

            return CallWindowProc( m_pListBoxHandler,
                                   Hwnd(),
                                   message,
                                   wParam,
                                   lParam);
        }

        case WM_MOVE:
        {
            if( m_fDragging )
            {
                m_fDragging = false;
                if( !IsParentMinimized() )
                {
                    RECT parentRect;
                    RECT rect;
                    GetWindowRect( Parent(), &parentRect );
                    GetWindowRect( Hwnd(), &rect );
                    int width = parentRect.right - parentRect.left;
                    int height = parentRect.bottom - parentRect.top;
                    int left;
                    int top;

                    UINT flags = SWP_NOZORDER | SWP_NOSIZE;


                    if( PosFlags() & e_fPosBindLeft )
                    {
                        left = rect.right;
                        top = rect.top;
                    }

                    if( PosFlags() & e_fPosBindRight )
                    {
                        left = rect.left - width;
                        top = rect.top;
                    }

                    ::SetWindowPos( Parent(),
                                    0,                // ignored
                                    left,
                                    top,
                                    width,
                                    height,
                                    flags );

                }
            }

            return CallWindowProc( m_pListBoxHandler,
                                   Hwnd(),
                                   message,
                                   wParam,
                                   lParam);
        }

        case WM_WINDOWPOSCHANGING:
        {
            LPWINDOWPOS pPos = (LPWINDOWPOS) lParam;
            if( Shrunk() )
            {
                // add 1 to maintain outline of caption
                // helps with 'gray-on-gray' situations
                pPos->cy = GetSystemMetrics( SM_CYSMCAPTION ) + 1;
            }
            return 0;
        }

        case MYWM_NOTIFYICON:
        {
            switch( lParam )
            {
                case WM_RBUTTONDOWN:
                {
                    switch( wParam )
                    {
                        case NOTIFY1:
                        {
                            if( m_fHidden )
                            {
                                ShowWindow( Hwnd(), SW_SHOW );
                                m_fHidden = false;
                            }
                            else
                            {
                                m_fHidden = true;
                                ShowWindow( Hwnd(), SW_HIDE );
                            }
                            break;
                        }
                    }
                    break;
                }

                case WM_MBUTTONDOWN:
                case WM_MBUTTONDBLCLK:
                {
                    switch( wParam )
                    {
                        case NOTIFY1:
                        {
                            ShrinkExpand();
                            if( m_fHidden )
                            {
                                ShowWindow( Hwnd(), SW_SHOW );
                                m_fHidden = false;
                            }


                            break;
                        }
                        break;
                    }
                    break;
                }

                case WM_LBUTTONDOWN:
                {
                    switch( wParam )
                    {
                        case NOTIFY1:
                        {
                            if( m_fHidden )
                            {
                                ShowWindow( Hwnd(), SW_SHOW );
                                m_fHidden = false;
                            }
                            if( IsParentMinimized() )
                            {
                                ShowWindow( Parent(), SW_RESTORE );
                                OnParentMove();
                            }
                            else
                            {
                                SetForegroundWindow( Parent() );
                            }
                            break;
                        }
                    }
                    break;
                }
            }
            return 0;
        }

        default:
            TRACE( "Message: " << hex << message );
            return CallWindowProc( m_pListBoxHandler,
                                   Hwnd(),
                                   message,
                                   wParam,
                                   lParam);
    }

    // unreachable
    return 0;
}

void SelectorList::OnChar( char ch )
{
    switch( ch )
    {
        case 13:            // Carriage Return
        case 32:            // Space bar
            ItemSelected( false );
            break;

        case 8:             // back space
        case 27:            // escape
            ResetSearch();
            SendMessage( Hwnd(), LB_SETCURSEL, -1, 0 );
            break;

        default:
        {
            m_searchStr += ch;
            SendMessage( Hwnd(), LB_SETCURSEL, -1, 0 );
            int idx = SendMessage( Hwnd(), LB_SELECTSTRING, -1, (LPARAM)(LPCSTR)m_searchStr.c_str() );
            if( idx == LB_ERR )
            {
                ResetSearch();
                TRACE( "Error from LB_SELECTSTRING: " << GetLastError() );
                return;
            }
            break;
        }
    }
}



void SelectorList::OnLButtonDown( WPARAM wParam )
{
    ResetSearch();
    if( SingleClickActivate() )
    {
        if( wParam & MK_CONTROL )
        {
            ItemSelected( true );
        }
        else
        {
            ItemSelected( false );
        }
    }
}

void SelectorList::OnLButtonDblClick( WPARAM wParam )
{
    ResetSearch();
    if( !SingleClickActivate() )
    {
        if( wParam & MK_CONTROL )
        {
            ItemSelected( true );
        }
        else
        {
            ItemSelected( false );
        }
    }
}

void SelectorList::OnRButtonDown( WPARAM wParam )
{
    ResetSearch();

    if( !ShrinkOnClick() )
    {
        return;
    }


    ShrinkExpand();

}

void SelectorList::ShrinkExpand()
{
    RECT rect;
    GetWindowRect( Hwnd(), &rect );

    if( !Shrunk() )
    {
        m_fShrunk = true;
        CopyMemory( &m_fullSizeRect, &rect, sizeof( RECT ));
        ::SetWindowPos( Hwnd(),
                      0,                // ignored
                      0,
                      0,
                      rect.right - rect.left,
                      0,
                      SWP_NOMOVE | SWP_NOZORDER );


        if( ShrinkTSE() )
        {
            RECT parentRect;
            GetWindowRect( Parent(), &parentRect );
            CopyMemory( &m_parentBigRect, &parentRect, sizeof( RECT ) );

            ::SetWindowPos( Parent(),
                            0,
                            0,
                            0,
                            parentRect.right - parentRect.left,
                            0,
                            SWP_NOMOVE | SWP_NOZORDER );
        }

    }
    else        // restore dimensions
    {
        m_fShrunk = false;
        SetRedraw( false );
        ::SetWindowPos( Hwnd(),
                      0,                // ignored
                      0,
                      0,
                      m_fullSizeRect.right - m_fullSizeRect.left,
                      m_fullSizeRect.bottom - m_fullSizeRect.top,
                      SWP_NOMOVE | SWP_NOZORDER );

        if( ShrinkTSE() )
        {
            ::SetWindowPos( Parent(),
                            0,
                            0,
                            0,
                            m_parentBigRect.right - m_parentBigRect.left,
                            m_parentBigRect.bottom - m_parentBigRect.top,
                            SWP_NOMOVE | SWP_NOZORDER );
        }

        SetRedraw( true );
    }
}

void SelectorList::ItemSelected( bool forRemoval )
{
    TRACE( "Item selected" );
    int choice = SendMessage( Hwnd(), LB_GETCURSEL, 0, 0 );

    if( LB_ERR == choice )
    {
        TRACE( "Error on LB_GETCURSEL" );
        ResetChoice();
        return;
    }

    SetChoice( choice, forRemoval );


//    SendMessage( Hwnd(), LB_SETCURSEL, -1, 0 );
    if( IsParentMinimized() )
    {
        ShowWindow( Parent(), SW_RESTORE );
    }
    OnParentMove();
}

void SelectorList::SetRedraw( bool enabled )
{
    TRACE( "Setting redraw: " << enabled );
    if( enabled && !Shrunk() && IsAutoSizing() && !IsParentMinimized() )
    {
        AutoSize();
    }
    SendMessage( Hwnd(), WM_SETREDRAW, enabled, 0 );
}

void SelectorList::ClearContent()
{
    SendMessage( Hwnd(), LB_RESETCONTENT, 0, 0 );
}

void SelectorList::InsertString( const char* szData )
{
    if( LB_ERR == SendMessage( Hwnd(),
                               LB_INSERTSTRING,
                               -1,
                               (LPARAM) (LPCSTR) szData ) )
    {
        TRACE( "Error inserting string: " << GetLastError() );
    }
}

void SelectorList::SetTitle( const char* szTitle )
{
    SetWindowText( Hwnd(), szTitle );
}

void SelectorList::SelectString( const char* szStr )
{
    int idx = SendMessage( Hwnd(), LB_SELECTSTRING, -1, (LPARAM)(LPCSTR)szStr );
    if( idx == LB_ERR )
    {
        TRACE( "Error setting current file: " << GetLastError() );
    }
}

void SelectorList::AutoSize()
{
    TRACE( "Autosizing" );
    // determine how big we can make window
    RECT desktopRect;
    GetWindowRect( GetDesktopWindow(), &desktopRect );

    RECT rect;
    GetWindowRect( Hwnd(), &rect );

    // grow downwards from current list box position
    LONG screenHeight = desktopRect.bottom - rect.top;

    // percent of screenHeight at which to stop growing
    const float threshold = 0.90;
    int maxHeight;

    if( PosFlags() & e_fPosBindLeft || PosFlags() & e_fPosBindRight )
    {
        RECT rectParent;
        GetWindowRect( Parent(), &rectParent );
        maxHeight = rectParent.bottom - rectParent.top;
    }
    else
    {
        maxHeight = screenHeight * threshold;
    }

    int items = SendMessage( Hwnd(), LB_GETCOUNT, 0, 0 );
    int itemHeight = SendMessage( Hwnd(), LB_GETITEMHEIGHT, 0, 0 );

    // leave a blank item at bottom to prevent scroll bar from showing unnecessarily
    items += 1;

    int listBoxHeight = (items * itemHeight);

    // add height of caption bar
    listBoxHeight += GetSystemMetrics( SM_CYSMCAPTION );

    if( listBoxHeight > maxHeight )
    {
        listBoxHeight = maxHeight;
    }

    TRACE( "Auto sizing" );
    ::SetWindowPos( Hwnd(),
                    0,                // ignored
                    rect.left,
                    rect.top,
                    rect.right - rect.left,
                    listBoxHeight,
                    SWP_NOZORDER );
}

void SelectorList::CalcDimensions( RECT& rect )
{
    int left = rect.left;
    int top = rect.top;
    int width = rect.right - rect.left;
    int height = rect.bottom - rect.top;

    TRACE( "Width: " << width );
    RECT desktopRect;
    GetWindowRect( GetDesktopWindow(), &desktopRect );

    int screenWidth = desktopRect.right - desktopRect.left;
    int screenHeight = desktopRect.bottom - desktopRect.top;

    RECT parentRect;
    GetWindowRect( Parent(), &parentRect );

    if( PosFlags() & e_fPosAbsolute )
    {
        if( PosFlags() & e_fAsPercent )
        {
            rect.left = (left / 100.0) * screenWidth;
            rect.top = (top / 100.0) * screenHeight;
            rect.right = rect.left + ((width / 100.0) * screenWidth);
            rect.bottom = rect.top + ((height / 100.0) * screenHeight);
        }
        else
        {
            rect.left = left;
            rect.top = top;
            rect.right = rect.left + width;
            rect.bottom = rect.top + height;
        }
        return;
    }

    if( PosFlags() & e_fPosBindLeft )
    {
        if( PosFlags() & e_fAsPercent )
        {
            rect.right = parentRect.left;
            rect.top = parentRect.top;
            rect.left = rect.right - ((width / 100.0) * screenWidth);
            rect.bottom = rect.top + ((height / 100.0) * screenHeight);
        }
        else
        {
            rect.right = parentRect.left;
            rect.top = parentRect.top;
            rect.left = parentRect.left - width;
            rect.bottom = rect.top + height;
            TRACE( "Left: " << rect.left );
        }
    }

    if( PosFlags() & e_fPosBindRight )
    {
        if( PosFlags() & e_fAsPercent )
        {
            rect.left = parentRect.right;
            rect.top = parentRect.top;
            rect.right = rect.left + ((width / 100.0) * screenWidth);
            rect.bottom = rect.top + ((height / 100.0) * screenHeight);
        }
        else
        {
            rect.left = parentRect.right;
            rect.top = parentRect.top;
            rect.right = rect.left + width;
            rect.bottom = rect.top + height;
        }
    }
}

void SelectorList::UpdatePosition( const RECT& rect )
{
    UINT flags = SWP_NOZORDER | SWP_NOSIZE;

    int left = rect.left;
    int top = rect.top;
    int width = rect.right - rect.left;
    int height = rect.bottom - rect.top;

    if( PosFlags() & e_fPosAbsolute )
    {
        return;
    }

    RECT parentRect;
    GetWindowRect( Parent(), &parentRect );

    if( PosFlags() & e_fPosBindLeft )
    {
        left = parentRect.left - width;
        top = parentRect.top;
    }

    if( PosFlags() & e_fPosBindRight )
    {
        left = parentRect.right;
        top = parentRect.top;
    }

    ::SetWindowPos( Hwnd(),
                    0,                // ignored
                    left,
                    top,
                    width,
                    height,
                    flags );
}

bool SelectorList::ParentMoved() const
{
    if( PosFlags() & e_fPosAbsolute )
    {
        return false;
    }

    if( IsParentMinimized() )
    {
        return false;
    }

    if( !NoDragging() )
    {
        RECT rect;
        GetWindowRect( Parent(), &rect );

        if( m_rectParent.left != rect.left || m_rectParent.top != rect.top )
        {
            return true;
        }
    }
    else
    {
        RECT parentRect;
        GetWindowRect( Parent(), &parentRect );

        RECT rect;
        GetWindowRect( Hwnd(), &rect );

        if( PosFlags() & e_fPosBindLeft )
        {
            if( parentRect.left != rect.right || parentRect.top != rect.top )
            {
                return true;
            }
            return false;
        }

        if( PosFlags() & e_fPosBindRight )
        {
            if( parentRect.right != rect.left || parentRect.top != rect.top )
            {
                return true;
            }
            return false;
        }

        return false;
    }
    return false;
}

void SelectorList::OnParentMove()
{
    TRACE( "*** Parent has moved" );
    GetWindowRect( Parent(),  &m_rectParent );
    RECT rect;
    GetWindowRect( Hwnd(), &rect );
    CalcDimensions( rect );
    UpdatePosition( rect );
}

bool SelectorList::IsParentMinimized() const
{
    WINDOWPLACEMENT wndPlacement;
    wndPlacement.length = sizeof( WINDOWPLACEMENT );

    GetWindowPlacement( Parent(), &wndPlacement );

    if( SW_SHOWMINIMIZED == wndPlacement.showCmd )
    {
        return true;
    }

    return false;
}

void SelectorList::SetChoice( int idx, bool forRemoval )
{
    CritSec::Lock lock( m_CritSec );
    m_iChoice = idx;
    m_fRemoveFile = forRemoval;
}

void SelectorList::ResetChoice()
{
    SetChoice( 0xFFFF, false );
}

bool SelectorList::Choice( int& idx, bool& forRemoval, bool resetChoice )
{
    CritSec::Lock lock( m_CritSec );
    if( 0xFFFF == m_iChoice )
    {
        return false;
    }

    idx = m_iChoice;
    forRemoval = m_fRemoveFile;

    if( resetChoice )
    {
        m_iChoice = 0xFFFF;
        m_fRemoveFile = false;
    }

    return true;
}

bool SelectorList::OnDragging()
{
    if( NoDragging() )
    {
        return false;
    }

    return true;
}
